
# DO NOT EDIT THIS FILE -- AUTOGENERATED BY PANTS
# Target: aio.core:package

from setuptools import setup

setup(**{
    'author': 'Ryan Northey',
    'author_email': 'ryan@synca.io',
    'description': 'A collection of functional utils for asyncio',
    'install_requires': (
        'abstracts>=0.0.12',
        'aiofiles',
        'aiohttp>=3.8.1',
        'orjson',
        'trycast>=0.7.3',
        'types-orjson',
    ),
    'license': 'Apache Software License 2.0',
    'long_description': """
aio.core
========

Utils for asyncio.
""",
    'maintainer': 'Ryan Northey',
    'maintainer_email': 'ryan@synca.io',
    'name': 'aio.core',
    'namespace_packages': (
    ),
    'package_data': {
        'aio.core': (
            'py.typed',
        ),
    },
    'packages': (
        'aio.core',
        'aio.core.dev',
        'aio.core.directory',
        'aio.core.directory.abstract',
        'aio.core.event',
        'aio.core.functional',
        'aio.core.log',
        'aio.core.stream',
        'aio.core.subprocess',
        'aio.core.tasks',
    ),
    'url': 'https://github.com/envoyproxy/pytooling/tree/main/aio.core',
    'version': '0.8.6',
})
